"""
Python package for running and visualizing the polarity model of organoid growth
"""

__version__ = '0.1.1'

# get main classes into top-level namespace
from .polarcore import Polar, PolarWNT, run_and_save
from .polargrid import PolarPDE

# get potential functions
from . import potentials
from . import potentials_wnt

# get plotting submodule
from . import plot


import torch


def potential(x, d, dx, lam_i, lam_j, pi, pj, qi, qj):
    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.sum(torch.cross(pi, qi, dim=2) *
                   torch.cross(pj, qj, dim=2), dim=2)
    S3 = torch.sum(torch.cross(qi, dx, dim=2) *
                   torch.cross(qj, dx, dim=2), dim=2)

    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    S = lam[:, :, 0] + lam[:, :, 1] * S1 + \
        lam[:, :, 2] * S2 + lam[:, :, 3] * S3
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij


def potential_nematic(x, d, dx, lam_i, lam_j, pi, pj, qi, qj):
    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.abs(torch.sum(torch.cross(pi, qi, dim=2)
                   * torch.cross(pj, qj, dim=2), dim=2))
    S3 = torch.abs(torch.sum(torch.cross(qi, dx, dim=2)
                   * torch.cross(qj, dx, dim=2), dim=2))

    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    S = lam[:, :, 0] + lam[:, :, 1] * S1 + \
        lam[:, :, 2] * S2 + lam[:, :, 3] * S3
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij


def potential_wnt_nematic(x, d, dx, lam_i, lam_j, pi, pj, qi, qj, wnt_cells, wnt_range):
    # displacements from each cell to each WNT-producing cell. shape: n_cells x n_wnt_cells x 3
    G = x[:, None, :] - x[wnt_cells]
    GG = G.clone()
    Gnorm = torch.sqrt(torch.sum(G**2, dim=2))
    GG[Gnorm > wnt_range] = 0
    GG = torch.nn.functional.normalize(GG, dim=2, eps=1e-6)

    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.abs(torch.sum(torch.cross(pi, qi, dim=2)
                   * torch.cross(pj, qj, dim=2), dim=2))
    S3 = torch.abs(torch.sum(torch.cross(qi, dx, dim=2)
                   * torch.cross(qj, dx, dim=2), dim=2))
    S4 = torch.sum(torch.sum(torch.cross(qi[:, 0, :][:, None, :].expand(
        qi.shape[0], G.shape[1], 3), GG, dim=2)**2, dim=2), dim=1)

    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    S = lam[:, :, 0] + lam[:, :, 1] * S1 + lam[:, :, 2] * S2 + lam[:, :,
                                                                   3] * S3 + lam[:, :, 4] * S4[:, None].expand(S4.shape[0], lam.shape[1])
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij


def potential_wnt(x, d, dx, lam_i, lam_j, pi, pj, qi, qj, wnt_cells, wnt_range):
    # displacements from each cell to each WNT-producing cell
    G = x[:, None, :] - x[wnt_cells]
    # displacements from each cell to each WNT-producing cell. shape: n_cells x n_wnt_cells x 3
    G = x[:, None, :] - x[wnt_cells]
    GG = G.clone()
    Gnorm = torch.sqrt(torch.sum(G**2, dim=2))
    GG[Gnorm > wnt_range] = 0
    GG = torch.nn.functional.normalize(GG, dim=2, eps=1e-6)

    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.sum(torch.cross(pi, qi, dim=2) *
                   torch.cross(pj, qj, dim=2), dim=2)
    S3 = torch.sum(torch.cross(qi, dx, dim=2) *
                   torch.cross(qj, dx, dim=2), dim=2)
    S4 = torch.sum(torch.sum(torch.cross(qi[:, 0, :][:, None, :].expand(
        qi.shape[0], G.shape[1], 3), G, dim=2)**2, dim=2), dim=1)

    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    S = lam[:, :, 0] + lam[:, :, 1] * S1 + lam[:, :, 2] * \
        S2 + lam[:, :, 3] * S3 + lam[:, :, 4] * S4
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij



import os
import plotly.graph_objects as go
from .plotcore import load, build_df, select

def plot(df):
    fig = go.Figure(data=[go.Cone(
        x=df['x1'],
        y=df['x2'],
        z=df['x3'],
        u=df['q1'],
        v=df['q2'],
        w=df['q3'],
        sizemode='absolute',
        sizeref=2,
    )])

    def fun(scene):
        scene.aspectmode = 'data'
        return
    fig.for_each_scene(fun)
    return fig

def save(fig, fname, T_plot):
    fig.write_html(os.path.join('animations', os.path.basename(fname)).replace('.pkl',f'_vectorfield_t_{T_plot}.html'),
                   include_plotlyjs='directory', full_html=False, animation_opts={'frame': {'duration': 100}})

if __name__ == "__main__":
    fname = input('Enter data filename (default: most recent): ')  or 'most recent'# 'data/test1.pkl'
    data, kwargs, fname = load(fname)
    T_plot = int(input('timestep to plot (default: last): ') or -1)
    df, _, kwargs = build_df(data, kwargs)
    df_t = select(df, T_plot, kwargs)
    fig = plot(df_t)
    save(fig, fname, T_plot)



import numpy as np
import math
import pickle

def init_grid(n, m, d = 2):
    """
    Initializes a square grid of cells at a distance of d cell radii (default = 2)

    AB polarity is initialized as 'up', and PCP is a random direction in the plane
    """
    xx = np.arange(n) - n/2 + 1/2
    yy = np.arange(m) - m/2 + 1/2
    idx = 0
    x = np.zeros((n*m, 3))
    p = np.zeros_like(x)
    q = np.zeros_like(x)
    Q = np.random.randn(3)
    Q = Q/np.sqrt((Q**2).sum())
    for X in xx:
        for Y in yy:
            x[idx, :] = [X*d, Y*d, 0]
            p[idx, :] = [0, 0, 1]
            q[idx, :] = Q
            idx += 1
    return x, p, q

def init_random_system(n, domain_size=30):
    """
    Initialize a random system with n cells. Position is normally distributed in each dimension, and each component of p and of q is uniformly distributed between -1 and 1
    
    Parameters
    ---------
    n : int
        Number of cells

    Returns
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell
    """
    x = domain_size * np.random.randn(n, 3)
    p = np.random.randn(n, 3)
    q = np.random.randn(n, 3)

    return x, p, q

def init_sphere(n, R = None):
    """
    Initialize a system with n points arranged in a sphere, with AB polarity pointing outward and PCP oriented randomly.

    To get x to lie on a sphere, sample x randomly and then normalize it.
    To get AB polarity to point outward, set it proportional to x

    Parameters
    ----------
    n : int
        Number of cells
    R : float, optional
        Radius of the sphere on which to place cells. If not specified, R is set so that the sphere's surface area is one per cell

    Returns
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell
    """

    if R is None:
        # set R so that on average each cell takes up unit area on the surface of the sphere
        R = np.sqrt(n/(4*np.pi))
    x = np.random.randn(n, 3)
    x = R*(x.T/np.linalg.norm(x, axis = 1)).T

    p = x/R
    q = np.random.randn(n, 3)

    return x, p, q

def init_tube(n, R = None, L = None, comb_direction = None):
    """
    Initializes a tube of radius R and length L.
    If either L or R is not specified, it is set so that each cell has a unit area on the tube.
    If neither L nor R is specified, L is set to 10*R

    AB polarity points out of the tube.
    PCP is oriented randomly.

    Parameters
    ----------
    n : int
        Number of cells
    R : float, optional
        Radius of the tube. Default: set so that each cell has a unit area on average
    L : float, optional
        Length of the tube. Default: 10*R
    comb_direction : str or None, optional
        direction in which to align q vectors, either along or around the tube.
        If None (default), leave q as random

    Returns
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell
    """
    if R is None and L is None:
        R = np.sqrt(n/(20*np.pi))
        L = 10*R
    elif L is None:
        L = n/(2*np.pi*R)
    elif R is None:
        R = n/(2*np.pi*L)

    # sample angle around tube and distance along the tube uniformly at random
    phi = 2*np.pi*np.random.rand(n)
    l = L*np.random.rand(n) - L/2
    # combine into 3D coordinates
    x = np.array([R*np.cos(phi), R*np.sin(phi), l]).T

    # make AB polarity point straight out of the tube and normalize to have length 1
    p = x.copy() / R
    p[:,2] = 0

    # initialize PCP at random
    q = 2*np.random.rand(*x.shape) - 1

    if comb_direction is not None:
        return comb_tube(x, p, q, direction = comb_direction)
    else:
        return x, p, q

def init_tube_grid(n, R = None, L = None, comb_direction = None):
    """
    Initializes a system with cells arranged on a tube, placed deterministically to be evenly spaced (i.e. on a grid)
    I'm doing this lazily so it may not end up with exactly n cells. Sorry.

    Parameters
    ----------
    n : int
        Number of cells
    R : float, optional
        Radius of the tube. Default: set so that each cell has a unit area on average
    L : float, optional
        Length of the tube. Default: 10*R
    comb_direction : str or None, optional
        direction in which to align q vectors, either along or around the tube.
        If None (default), leave q as random

    Returns
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell
    """
    if R is None and L is None:
        R = np.sqrt(n/(20*np.pi))
        L = 10*R
    elif L is None:
        L = n/(2*np.pi*R)
    elif R is None:
        R = n/(2*np.pi*L)

    # generate distance along the tube and angle around the tube evenly spaced
    l = 0.5 + np.arange(-L/2, L/2)
    phi = 2*np.pi*np.linspace(0, 1, int(n/L), endpoint = False)

    # combine into 3D coordinates
    x = np.empty((len(l)*len(phi), 3), dtype = float)
    idx = 0
    dphi = np.pi / len(phi)
    for ph in phi:
        for j, ell in enumerate(l):
            x[idx, :] = np.array([R*np.cos(ph + j*dphi), R*np.sin(ph + j*dphi), ell])
            idx+=1

    # make AB polarity point straight out of the tube and normalize to have length 1
    p = x.copy() / R
    p[:,2] = 0

    # initialize PCP, either around the tube or at random
    q = 2*np.random.rand(*x.shape) - 1

    if comb_direction is not None:
        return comb_tube(x, p, q, direction = comb_direction)
    else:
        return x, p, q

def comb_tube(x, p, q, direction = 'around'):
    """
    align the PCP vectors (q) to point around the tube on which cells are positioned (x).

    This function assumes that:
        x.shape = (n, 3) and each x[i,:] lies on a tube centered on the z-axis
        p[:, :2] = x[:, :2] and p[:, 2] = 0

    Parameters:
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell

    Returns
    ----------
    x : np.ndarray
        Position of each cell in 3D space
    p : np.ndarray
        AB polarity vector of each cell
    q : np.ndarray
        PCP vector of each cell
    """
    if direction == 'around':
        q = np.vstack((p[:,1], -p[:,0], p[:,2])).T
    elif direction == 'along':
        q = np.zeros_like(q)
        q[:, 2] = 1
    else:
        raise NotImplementedError('direction must be either \'around\' or \'along\'')
    return x, p, q

def comb_sphere(x, p, q):
    q = np.vstack((p[:,1], -p[:,0], np.zeros_like(p[:,2]))).T
    return x, p, q

def load_cached(fname):
    with open(fname, 'rb') as fobj:
        x, p, q = pickle.load(fobj)
    return x, p, q


def init_square_system(n, domain_size=30):
    x = domain_size * np.random.random((n, 3))
    p = np.random.randn(n, 3)
    q = np.random.randn(n, 3)

    return x, p, q

def init_random_rectangle(n, domain_size1=30, domain_size2=50):
    x = np.random.random((n, 3))
    x[:, :2] *= domain_size1
    x[:, 2] *= domain_size2

    p = np.random.randn(n, 3)
    q = np.random.randn(n, 3)

    return x, p, q

def fibonacci_sphere(samples=1, radius = 30):
    rnd = 1.
    points = []
    offset = 2./samples
    increment = math.pi * (3. - math.sqrt(5.))

    for i in range(samples):
        y = ((i * offset) - 1) + (offset / 2)
        r = math.sqrt(1 - pow(y,2))

        phi = ((i + rnd) % samples) * increment

        x = math.cos(phi) * r
        z = math.sin(phi) * r

        points.append([x,y,z])
    p = np.array(points)
    x = p.copy() * radius
    q = np.random.randn(samples, 3)    
    return x, p, q

def init_cylinder(n, domain_size=30, ratio=50):
    phi = 2 * np.pi * np.random.random(n)
    x = domain_size/ratio * np.cos(phi) + domain_size/2
    y = domain_size/ratio * np.sin(phi) + domain_size/2
    z = 0.99999 * domain_size * np.random.random(n)

    x = np.array([x, y, z]).T

    p = x.copy()
    p -= np.mean(p, axis=0)
    p[:, 2] = 0
    q = np.cross(p, [0, 0, 1])

    return x, p, q

def init_solid_cylinder(n, domain_size=30, ratio=50):
    phi = 2 * np.pi * np.random.random(n)
    r = np.sqrt(np.random.random(n))
    x = r * domain_size/ratio * np.cos(phi) + domain_size/2
    y = r * domain_size/ratio * np.sin(phi) + domain_size/2
    z = 0.99999 * domain_size * np.random.random(n)

    x = np.array([x, y, z]).T

    p = x.copy()
    p -= np.mean(p, axis=0)
    p[:, 2] = 0
    q = np.cross(p, [0, 0, 1])

    return x, p, q

def init_torus(n, domain_size=30, ratio=5):
    phi = 2 * np.pi * np.random.random(n)

    x0 = domain_size * np.cos(phi)
    y0 = domain_size * np.sin(phi)

    phi2 = 2 * np.pi * np.random.random(n)

    x = x0 + np.cos(phi) * domain_size / ratio * np.cos(phi2)
    y = y0 + np.sin(phi) * domain_size / ratio * np.cos(phi2)
    z = domain_size / ratio * np.sin(phi2)

    x = np.array([x, y, z]).T

    p = x.copy()
    p[:, 0] -= x0
    p[:, 1] -= y0

    c = x.copy()
    c[:, 0] = y0
    c[:, 1] = -x0
    c[:, 2] *= 0

    p /= np.sqrt(np.sum(p**2))
    c /= np.sqrt(np.sum(c**2))
    q = np.cross(p, c)

    return x, p, q

def init_torus_cross(n, domain_size=30, ratio=5):

    x, p, q = init_torus(n, domain_size=domain_size, ratio=ratio)

    xc, pc, qc = init_cylinder(int(n/np.pi), domain_size=0.8 * 2*domain_size, ratio=ratio*2)

    xc[:, 1], xc[:, 2] = xc[:, 2].copy(), xc[:, 1].copy()
    pc[:, 1], pc[:, 2] = pc[:, 2].copy(), pc[:, 1].copy()
    qc[:, 1], qc[:, 2] = qc[:, 2].copy(), qc[:, 1].copy()

    xc -= np.mean(xc, axis=0)

    return np.concatenate((x, xc)), np.concatenate((p, pc)), np.concatenate((q, qc))

def init_plane(n, L = 25):
    x = np.zeros(n)
    y = L * np.random.random(n)
    z = L * np.random.random(n)

    x = np.array([x, y, z]).T
    x -= np.mean(x, axis=0)

    d2 = np.sort(np.sum((x[:, None, :] - x[None, :, :])**2, axis=2), axis=1)
    mean_dist = np.mean(np.sqrt(d2[:, 1:5]))
    x *= np.sqrt(2)/mean_dist

    p = np.zeros_like(x)
    p[:, 0] = 1
    q = np.cross(p, [0, 0, 1])

    return x, p, q


"""
This script should open a data file in the legacy format and save it in a forward-compatible format
"""
from .plot import plotcore as pc
import pickle as pkl

def convert(fname):
    data, kwargs, fname = pc.load(fname)
    try:
        df, lig, kwargs = pc.build_df(data, kwargs)
        fname_new = fname.replace('data','data_converted')
        mdict = {'df':df, 'lig':lig, 'kwargs':kwargs}
        with open(fname_new, 'wb') as fobj:
            pkl.dump(mdict, fobj)
        return
    except ValueError:
        pass


if __name__ == '__main__':
    fname = input('file name? (default: most recent) ') or 'most recent'
    convert(fname)


import numpy as np
import torch
from scipy.spatial.kdtree import KDTree
import torch.linalg as la
import itertools
import time
import pickle
from .plot.plotcore import build_df
# import time


# http://stackoverflow.com/questions/14016898/port-matlab-bounding-ellipsoid-code-to-python
# http://stackoverflow.com/questions/1768197/bounding-ellipse/1768440#1768440
# https://minillinim.github.io/GroopM/dev_docs/groopm.ellipsoid-pysrc.html

def dd_factory(start_time, rate, max_cells):
    def division_decider(sim, tstep):
        """
        This is a function that decides whether or not to let the cells divide

        Idea: take a sublinear function of time, and allow cell division whenever the value of that function passes an integer
        This will make cell division happen more rarely as the simulation progresses.
        """
        T = sim.dt * tstep
        if T < start_time or len(sim.x) > max_cells - 1:
            return False

        def f(T): return rate*T
        if int(f(T)) > int(f(T-sim.dt)):
            return True
        else:
            return False
    return division_decider

def mvee(points, tol=0.0001):
    """
    Finds the ellipse equation in "center form"
    (x-c).T * A * (x-c) = 1
    """
    device = torch.device('cuda')
    dot = lambda foo, bar : torch.tensordot(foo, bar, dims=1)
    N, d = points.shape
    Q = torch.column_stack((points, torch.ones(N, device=device))).T
    err = tol+1.0
    u = torch.ones(N, device = device)/N
    while err > tol:
        # assert u.sum() == 1 # invariant
        X = dot(dot(Q, torch.diag(u)), Q.T)
        M = torch.diag(dot(dot(Q.T, la.inv(X)), Q))
        jdx = torch.argmax(M)
        step_size = (M[jdx]-d-1.0)/((d+1)*(M[jdx]-1.0))
        new_u = (1-step_size)*u
        new_u[jdx] += step_size
        err = la.norm(new_u-u)
        u = new_u
    c = dot(u, points)
    A = la.inv(dot(dot(points.T, torch.diag(u)), points)
               - torch.outer(c, c))/d
    return A, c

class Polar:
    """
    Class to define and run simulations of the polarity model of cell movement

    Examples:
    ----------
        ```
        sim = Polar(x, p, q, lam, beta, eta=eta, yield_every=yield_every, init_k=50)
        runner = sim.simulation(potential=potential)

        # Running the simulation
        data = []  # For storing data
        i = 0
        t1 = time.time()
        print('Starting')

        for xx, pp, qq, lam in itertools.islice(runner, timesteps):
            i += 1
            print(f'Running {i} of {timesteps}   ({yield_every * i} of {yield_every * timesteps})   ({len(xx)} cells)')
            data.append((xx, pp, qq, lam))

            if len(xx) > 1000:
                print('Stopping')
                break
        ```
    """
    def __init__(self, x, p, q, lam, beta, eta, yield_every, dt = 0.1, beta_decay = 1.0, do_nothing_threshold = 1e-5, divide_single = False,
                 device='cuda', dtype=torch.float, init_k=100, callback=None):
        self.device = device
        self.dtype = dtype

        self.k = init_k
        self.true_neighbour_max = init_k//2
        self.d = None
        self.idx = None
        self.callback = callback

        self.x = x
        self.p = p
        self.q = q
        self.beta = beta
        self.dt = dt
        self.sqrt_dt = None
        self.lam = lam

        self.init_simulation()
        self.eta = eta
        self.yield_every = yield_every
        self.beta_decay = beta_decay
        self.do_nothing_threshold = do_nothing_threshold
        self.divide_single = divide_single

    def init_simulation(self):
        """
        Checks input dimensions, cleans and converts into torch.Tensor types

        Parameters
        ----------
            dt : float
                size of time step for simulation
            lam : array
                weights for the terms of the potential function, possibly different for each cell.
            p : array_like
                AB polarity vector of each cell
            q : array_like
                PCP vector of each cell
            x : array_like
                Position of each cell in 3D space
            beta : array_like
                for each cell, probability of division per unit time
            kwargs : dict
                keyword arguments

        Returns
        ----------
            None
        """
        assert len(self.x) == len(self.p)
        assert len(self.q) == len(self.x)
        assert len(self.beta) == len(self.x)

        sqrt_dt = np.sqrt(self.dt)
        x = torch.tensor(self.x, requires_grad=True, dtype=self.dtype, device=self.device)
        p = torch.tensor(self.p, requires_grad=True, dtype=self.dtype, device=self.device)
        q = torch.tensor(self.q, requires_grad=True, dtype=self.dtype, device=self.device)

        beta = torch.tensor(self.beta, dtype=self.dtype, device=self.device)

        lam = torch.tensor(self.lam, dtype=self.dtype, device=self.device)
        # if lam is not given per-cell, return an expanded view
        if len(lam.shape) == 1:
            lam = lam.expand(x.shape[0], lam.shape[0]).clone()

        self.x = x
        self.p = p
        self.q = q
        self.beta = beta
        self.sqrt_dt = sqrt_dt
        self.lam = lam
        return

    def find_potential_neighbours(self, k=None, distance_upper_bound=np.inf, workers = -1):
        """
        Uses cKDTree to compute potential nearest-neighbors of each cell

        Parameters
        ----------
            x : array_like
                Position of each cell in 3D space
            k : list of integer or integer
                The list of k-th nearest neighbors to return. If k is an integer it is treated as a list of [1, ... k] (range(1, k+1)). Note that the counting starts from 1.
            distance_upper_bound : nonnegative float, optional
                Return only neighbors within this distance. This is used to prune tree searches, so if you are doing a series of nearest-neighbor queries, it may help to supply the distance to the nearest neighbor of the most recent point. Default: np.inf
            workers: int, optional
                Number of workers to use for parallel processing. If -1 is given, all CPU threads are used. Default: -1.

        Returns
        ----------
            d : array
                distance from each cell to each of its potential neighbors
            idx : array
                index of each cell's potential neighbors
        """
        if k is None:
            k = self.k
        x = self.x.detach().to('cpu').numpy()
        tree = KDTree(x)
        d, idx = tree.query(x, k + 1, distance_upper_bound=distance_upper_bound, workers=workers)
        self.d = torch.tensor(d[:, 1:], dtype=torch.long, device=self.device)
        self.idx = torch.tensor(idx[:, 1:], dtype=torch.long, device=self.device)
        
    def find_true_neighbours(self):
        """
        Finds the true neighbors of each cell

        Parameters
        ----------
            d : array
                distance from each cell to each of its potential neighbors
            dx : array
                displacement vector from each cell to each of its potential neighbors

        Returns
        ----------
            z_mask : torch.tensor
        """
        full_n_list = self.x[self.idx]
        self.dx = self.x[:, None, :] - full_n_list
        with torch.no_grad():
            z_masks = []
            i0 = 0
            batch_size = 250
            i1 = batch_size
            while True:
                if i0 >= self.dx.shape[0]:
                    break
                # ?
                n_dis = torch.sum((self.dx[i0:i1, :, None, :] / 2 - self.dx[i0:i1, None, :, :]) ** 2, dim=3)
                # ??
                n_dis += 100000 * torch.eye(n_dis.shape[1], device=self.device, dtype=self.dtype)[None, :, :]

                z_mask = torch.sum(n_dis < (self.d[i0:i1, :, None] ** 2 / 4), dim=2) <= 0
                z_masks.append(z_mask)

                if i1 > self.dx.shape[0]:
                    break
                i0 = i1
                i1 += batch_size
        z_mask = torch.cat(z_masks, dim=0)
        sort_idx = torch.argsort(z_mask.int(), dim=1, descending=True)

        z_mask = torch.gather(z_mask, 1, sort_idx)
        dx = torch.gather(self.dx, 1, sort_idx[:, :, None].expand(-1, -1, 3))
        idx = torch.gather(self.idx, 1, sort_idx)

        m = torch.max(torch.sum(z_mask, dim=1))
        m = min(m, z_mask.shape[1])

        self.z_mask = z_mask[:, :m]
        self.dx = dx[:, :m]
        self.idx = idx[:, :m]
        # there is an issue here. it assumes that every cell has the same # of true neighbors
        # moreover it typically overestimates the number of true neighbors, apparently
        self.d = torch.linalg.norm(self.dx, dim = -1)
        return m

    def potential(self, potential):
        """
        Computes the potential energy of the system
        
        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            idx : array_like
                indices of potential nearest-neighbors of each cell
            d : array_like
                distances from each cell to each of the potential nearest-neighbors specified by idx
            lam : array
                array of weights for the terms that make up the potential
            potential : callable
                function that computes the value of the potential between two cells, i and j
                call signature (x, d, dx, lam_i, lam_j, pi, pj, qi, qj, **kwargs)
            kwargs : dict
                keyword arguments, passed to the potential function
        
        Returns
        ----------
            V : torch.Tensor
                value of the potential
            m : int
                largest number of true neighbors of any cell
        """
        # Find neighbours
        m = self.find_true_neighbours() # return value = maximum number of true neighbors of any cell
        
        # Normalise dx
        dx = self.dx / torch.linalg.norm(self.dx, dim = -1)[:, :, None]

        # Calculate S
        pi = self.p[:, None, :].expand(self.p.shape[0], self.idx.shape[1], 3)
        pj = self.p[self.idx]
        qi = self.q[:, None, :].expand(self.q.shape[0], self.idx.shape[1], 3)
        qj = self.q[self.idx]

        lam_i = self.lam[:, None, :].expand(self.p.shape[0], self.idx.shape[1], self.lam.shape[1])
        lam_j = self.lam[self.idx]

        Vij = potential(self.x, self.d, dx, lam_i, lam_j, pi, pj, qi, qj)
        V = torch.sum(self.z_mask.float() * Vij)

        return V, int(m)

    def update_k(self, true_neighbour_max, tstep):
        """
        Dynamically adjusts the number of neighbors to look for.

        If very few of the potential neighbors turned out to be true, you can look for fewer potential neighbors next time.
        If very many of the potential neighbors turned out to be true, you should look for more potential neighbors next time.

        Parameters
        ----------
            true_neighbor_max : int
                largest number of true neighbors of any cell found most recently
            tstep : int
                how many time steps of simulation have elapsed

        Returns
        ----------
            k : int
                new max number of potential neighbors to seek
            n_update : int
                controls when to next check for potential neighbors
        """
        k = self.k
        fraction = true_neighbour_max / k
        if fraction < 0.25:
            k = int(0.75 * k)
        elif fraction > 0.75:
            k = int(1.5 * k)
        n_update = 1 if tstep < 50 else max([1, int(20 * np.tanh(tstep / 200))])
        self.k = k
        return n_update

    def gradient_step(self, tstep, potential):
        """
        Move the simulation forward by one time step

        Parameters
        ----------
            dt : float
                Size of the time step
            eta : float
                Strength of the added noise
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            sqrt_dt : float
                square root of dt. To be used for normalizing the size of the noise added per time step
            tstep : int
                how many simulation time steps have elapsed
            x : torch.Tensor
                Position of each cell in 3D space
            potential : callable
                function that computes the value of the potential between two cells, i and j
                call signature (x, d, dx, lam_i, lam_j, pi, pj, qi, qj, **kwargs)
            kwargs : dict
                keyword arguments, to be passed to the potential function

        Returns
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
        """
        # Normalise p, q
        with torch.no_grad():
            self.p /= torch.sqrt(torch.sum(self.p ** 2, dim=1))[:, None]
            self.q /= torch.sqrt(torch.sum(self.q ** 2, dim=1))[:, None]

        # Calculate potential
        V, self.true_neighbour_max = self.potential(potential=potential)

        # Backpropagation
        V.backward()

        # Time-step
        with torch.no_grad():
            self.x += -self.x.grad * self.dt + self.eta * torch.empty(*self.x.shape, dtype=self.dtype, device=self.device).normal_() * self.sqrt_dt
            self.p += -self.p.grad * self.dt + self.eta * torch.empty(*self.x.shape, dtype=self.dtype, device=self.device).normal_() * self.sqrt_dt
            self.q += -self.q.grad * self.dt + self.eta * torch.empty(*self.x.shape, dtype=self.dtype, device=self.device).normal_() * self.sqrt_dt

            if self.callback is not None:
                self.callback(tstep * self.dt, self.x, self.p, self.q, self.lam)

        # Zero gradients
        self.x.grad.zero_()
        self.p.grad.zero_()
        self.q.grad.zero_()

        return

    def simulation(self, potential, division_decider = lambda *args : True):
        """
        Generator to implement the simulation

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            eta : float
                Strength of the added noise
            potential : callable
                function that computes the value of the potential between two cells, i and j
                call signature (x, d, dx, lam_i, lam_j, pi, pj, qi, qj)
            yield_every : int, optional
                How many simulation time steps to take between yielding the system state. Default: 1
            dt : float, optional
                Size of the time step. Default: 0.1
            kwargs : dict
                Keyword args passed to self.init_simulation and self.time_step.
                Values passed here override default values
                dt : float
                    time step
                yield_every : int
                    how many time steps to take in between yielding system state

        Yields
        ----------
            x : numpy.ndarray
                Position of each cell in 3D space
            p : numpy.ndarray
                AB polarity vector of each cell
            q : numpy.ndarray
                PCP vector of each cell
            lam : numpy.ndarray
                weights for the terms of the potential function for each cell.
        """

        tstep = 0
        while True:
            tstep += 1

            # perform cell division, depending on the output of the function division_decider
            # by default, always do cell division (this results in exponential growth of number of cells)
            division = False
            if division_decider(self, tstep):
                if self.divide_single:
                    division = self.cell_division_single()
                else:
                    division = self.cell_division()
            
            n_update = self.update_k(self.true_neighbour_max, tstep)
            self.k = min(self.k, len(self.x) - 1)

            if division or tstep % n_update == 0 or self.idx is None:
                self.find_potential_neighbours()

            

            self.gradient_step(tstep, potential=potential)

            if tstep % self.yield_every == 0:
                xx = self.x.detach().to("cpu").numpy().copy()
                pp = self.p.detach().to("cpu").numpy().copy()
                qq = self.q.detach().to("cpu").numpy().copy()
                ll = self.lam.detach().to("cpu").numpy().copy()
                yield xx, pp, qq, ll

    def cell_division(self):
        """
        Decides which cells divide, and if they do, places daughter cells.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            dt : float
                Size of the time step.
            kwargs : dict
                Valid keyword arguments:
                beta_decay : float
                    the factor by which beta (probability of cell division per unit time) decays upon cell division.
                    after cell division, one daughter cell has the same beta as the mother (b0), and the other has beta = b0 * beta_decay

        Returns
        ---------
            division : bool
                True if cell division has taken place, otherwise False
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False
        
        # set probability according to beta and dt
        d_prob = self.beta * self.dt
        # flip coins
        draw = torch.empty_like(self.beta).uniform_()
        # find successes
        events = draw < d_prob
        division = False

        if torch.sum(events) > 0:
            with torch.no_grad():
                division = True
                # find cells that will divide
                idx = torch.nonzero(events)[:, 0]

                x0 = self.x[idx, :]
                p0 = self.p[idx, :]
                q0 = self.q[idx, :]
                l0 = self.lam[idx, :]
                b0 = self.beta[idx] * self.beta_decay

                # make a random vector and normalize to get a random direction
                move = torch.empty_like(x0).normal_()
                move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

                # move the cells so that the center of mass of each pair is at the same place as the mother cell was
                x0 = x0 - move/2
                self.x[idx, :] += move/2

                # append new cell data to the system state
                self.x = torch.cat((self.x, x0))
                self.p = torch.cat((self.p, p0))
                self.q = torch.cat((self.q, q0))
                self.lam = torch.cat((self.lam, l0))
                self.beta = torch.cat((self.beta, b0))

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return division

    def cell_division_single(self):
        """
        Selects exactly one cell to divide and divides it.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False

        idx = torch.multinomial(self.beta, 1)
        with torch.no_grad():
            x0 = self.x[idx, :]
            p0 = self.p[idx, :]
            q0 = self.q[idx, :]
            l0 = self.lam[idx, :]
            b0 = self.beta[idx] * self.beta_decay

            # make a random vector and normalize to get a random direction
            move = torch.empty_like(x0).normal_()
            move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

            # move the cells so that the center of mass of each pair is at the same place as the mother cell was
            x0 = x0 - move/2
            self.x[idx, :] += move/2

            # append new cell data to the system state
            self.x = torch.cat((self.x, x0))
            self.p = torch.cat((self.p, p0))
            self.q = torch.cat((self.q, q0))
            self.lam = torch.cat((self.lam, l0))
            self.beta = torch.cat((self.beta, b0))

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return True

class PolarWNT(Polar):
    def __init__(self, *args, wnt_cells = None, wnt_threshold = 1e-2, wnt_decay = 0, **kwargs):
        self.wnt_cells = wnt_cells
        self.wnt_threshold = wnt_threshold
        self.wnt_decay = wnt_decay
        super().__init__(*args, **kwargs)
        # initialize a G (WNT gradient) vector as zeros
        self.G = torch.zeros_like(self.x)

    def init_simulation(self):
        """
        Checks input dimensions, cleans and converts into torch.Tensor types
        If there are WNT cells, it adds WNT to them in the w tensor

        Parameters
        ----------
            dt : float
                size of time step for simulation
            lam : array
                weights for the terms of the potential function, possibly different for each cell.
            p : array_like
                AB polarity vector of each cell
            q : array_like
                PCP vector of each cell
            w : array_like
                WNT concentration at each cell
            x : array_like
                Position of each cell in 3D space
            beta : array_like
                for each cell, probability of division per unit time
            kwargs : dict
                keyword arguments

        Returns
        ----------
            None
        """
        assert len(self.x) == len(self.p)
        assert len(self.q) == len(self.x)
        assert len(self.beta) == len(self.x)

        sqrt_dt = np.sqrt(self.dt)
        x = torch.tensor(self.x, requires_grad=True, dtype=self.dtype, device=self.device)
        p = torch.tensor(self.p, requires_grad=True, dtype=self.dtype, device=self.device)
        q = torch.tensor(self.q, requires_grad=True, dtype=self.dtype, device=self.device)
        w = torch.zeros(len(x), requires_grad=False, dtype=self.dtype, device=self.device)
        if self.wnt_cells is not None:
            w[self.wnt_cells] = 1

        beta = torch.tensor(self.beta, dtype=self.dtype, device=self.device)

        lam = torch.tensor(self.lam, dtype=self.dtype, device=self.device)
        # if lam is not given per-cell, return an expanded view
        if len(lam.shape) == 1:
            lam = lam.expand(x.shape[0], lam.shape[0]).clone()

        self.x = x
        self.p = p
        self.q = q
        self.w = w
        self.beta = beta
        self.sqrt_dt = sqrt_dt
        self.lam = lam
        return

    def get_gradient_vectors_better(self):
        self.find_true_neighbours()
        dx = self.dx.clone()
        d = self.d.clone()
        with torch.no_grad():
            w_mask = self.w >= self.wnt_threshold
            # Calculate weights tensors
            w_ii = self.w[:, None].expand(self.w.shape[0], self.idx.shape[1])
            w_ij = self.w[self.idx]
            Gij = (-w_ii[:, :, None] + w_ij[:, :, None]) * dx / d[:,:,None]**2
            G = torch.mean(self.z_mask[:, :, None].float() * Gij, dim=1)
            d_G = torch.sqrt(torch.sum(G**2, dim =1))
            cells_to_normalize = torch.logical_and(w_mask, d_G > 1e-3)
            G[cells_to_normalize] /= d_G[cells_to_normalize, None]
            G[d_G <= 1e-3] = torch.zeros(3, dtype=G.dtype, device=G.device)
            #G[~weighted_cells]= 0
        self.G = G
        return G, self.w

    def get_gradient_vectors(self, better = False):
        if better:
            return self.get_gradient_vectors_better()
        # Compute gradient vectors
        self.find_true_neighbours()
        dx = self.dx.clone()
        with torch.no_grad():

            # Count Paneth cells as weighted cells only inside this function to establish gradient
            weighted_cells = torch.where(self.w >= self.wnt_threshold)[0]

            # Calculate weights tensors
            w_ii = self.w[:, None].expand(self.w.shape[0], self.idx.shape[1])
            w_ij = torch.zeros_like(w_ii)       # w_ij is zero for all non-weighted cells i
            w_ij[weighted_cells, :] = self.w[self.idx[weighted_cells]] # if cell i has w>threshold, then the i-th row of w_ij will have the w-values of all of its neighbors

            # Normalize dx compute G
            d_dx = torch.sqrt(torch.sum(dx ** 2, dim=2))
            dx /= d_dx[:, :, None]              # dx[i,j,:] is the unit vector pointing from cell i to cell j. shape is (N, idx.shape[1], 3)
            Gij = (w_ii[:, :, None] + w_ij[:, :, None]) * dx
            G = torch.sum(self.z_mask[:, :, None].float() * Gij, dim=1)

            # Project G vectors so divisions occur in cell-cell-plane and normalize
            G_tilde = -torch.cross(torch.cross(G, self.p), self.p)  # Note the minus!
            # G_tilde = G
            d_G_tilde = torch.sqrt(torch.sum(G_tilde ** 2, dim=1))
            # find cells whose d_G_tilde is below some threshold, and set those cells' G to zero
            # if you don't do this you'll get NANs
            G_tilde[d_G_tilde < 1e-3] = 0
            G_tilde[weighted_cells] /= d_G_tilde[weighted_cells, None]
        self.G = G_tilde
        return G_tilde, self.w
    
    def get_gradient_averaging(self):
        # Smoothening the WNT gradient based on true local neighbourhood
        self.find_true_neighbours()
        w_copy = self.w.clone()

        with torch.no_grad():
            weights_true_neighb = self.w[self.idx] * self.z_mask.float()
            w_copy = (torch.sum(weights_true_neighb, dim=1) + w_copy) / (torch.sum(self.z_mask, dim=1) + 1).float()
        
        self.w = w_copy
        # replenish WNT at source cells
        self.w[self.wnt_cells] = 1
        return w_copy

    def potential(self, potential):
        # Find neighbours
        m = self.find_true_neighbours() # return value = maximum number of true neighbors of any cell
        
        # Normalise dx
        dx = self.dx / torch.linalg.norm(self.dx, dim = -1)[:, :, None]

        # Normalise p, q
        with torch.no_grad():
            self.p /= torch.sqrt(torch.sum(self.p ** 2, dim=1))[:, None]
            self.q /= torch.sqrt(torch.sum(self.q ** 2, dim=1))[:, None]

        # Calculate S in the following
        pi = self.p[:, None, :].expand(self.p.shape[0], self.idx.shape[1], 3)
        pj = self.p[self.idx]
        qi = self.q[:, None, :].expand(self.q.shape[0], self.idx.shape[1], 3)
        qj = self.q[self.idx]
        Gi = self.G[:, None, :].expand(self.G.shape[0], self.idx.shape[1], 3)

        lam_i = self.lam[:, None, :].expand(self.p.shape[0], self.idx.shape[1], self.lam.shape[1])
        lam_j = self.lam[self.idx]
        Vij = potential(self.x, self.d, dx, lam_i, lam_j, pi, pj, qi, qj, Gi)
        V = torch.sum(self.z_mask.float() * Vij)

        return V, int(m)

    def simulation(self, potential, division_decider = lambda *args : True, better_WNT_gradient = False):
        """
        Generator to implement the simulation

        Note: you can interact with this thing. Example:
        ```python
        polarguy = PolarWNT(**)
        sim = polarguy.simulation(***)
        for out in sim:
            data.append(out)
            polarguy.get_gradient_averaging()
        ```

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            eta : float
                Strength of the added noise
            potential : callable
                function that computes the value of the potential between two cells, i and j
                call signature (x, d, dx, lam_i, lam_j, pi, pj, qi, qj)
            yield_every : int, optional
                How many simulation time steps to take between yielding the system state. Default: 1
            dt : float, optional
                Size of the time step. Default: 0.1
            kwargs : dict
                Keyword args passed to self.init_simulation and self.time_step.
                Values passed here override default values
                dt : float
                    time step
                yield_every : int
                    how many time steps to take in between yielding system state

        Yields
        ----------
            x : numpy.ndarray
                Position of each cell in 3D space
            p : numpy.ndarray
                AB polarity vector of each cell
            q : numpy.ndarray
                PCP vector of each cell
            w : numpy.ndarray
                WNT concentration at each cell
            lam : numpy.ndarray
                weights for the terms of the potential function for each cell.
        """

        tstep = 0
        while True:
            tstep += 1

            # perform cell division, depending on the output of the function division_decider
            # by default, always do cell division (this results in exponential growth of number of cells)
            division = False
            if division_decider(self, tstep):
                if self.divide_single:
                    division = self.cell_division_single()
                else:
                    division = self.cell_division()
            
            n_update = self.update_k(self.true_neighbour_max, tstep)
            self.k = min(self.k, len(self.x) - 1)

            if division or tstep % n_update == 0 or self.idx is None:
                self.find_potential_neighbours()
            
            self.get_gradient_vectors(better=better_WNT_gradient)
            self.gradient_step(tstep, potential=potential)
            self.w = self.w * np.exp(self.dt * self.wnt_decay)

            if tstep % self.yield_every == 0:
                xx = self.x.detach().to("cpu").numpy().copy()
                pp = self.p.detach().to("cpu").numpy().copy()
                qq = self.q.detach().to("cpu").numpy().copy()
                ww = self.w.detach().to("cpu").numpy().copy()
                ll = self.lam.detach().to("cpu").numpy().copy()
                yield xx, pp, qq, ww, ll
    def cell_division(self):
        """
        Decides which cells divide, and if they do, places daughter cells.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            dt : float
                Size of the time step.
            kwargs : dict
                Valid keyword arguments:
                beta_decay : float
                    the factor by which beta (probability of cell division per unit time) decays upon cell division.
                    after cell division, one daughter cell has the same beta as the mother (b0), and the other has beta = b0 * beta_decay

        Returns
        ---------
            division : bool
                True if cell division has taken place, otherwise False
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False
        
        # set probability according to beta and dt
        d_prob = self.beta * self.dt
        # flip coins
        draw = torch.empty_like(self.beta).uniform_()
        # find successes
        events = draw < d_prob
        division = False

        if torch.sum(events) > 0:
            with torch.no_grad():
                division = True
                # find cells that will divide
                idx = torch.nonzero(events)[:, 0]

                x0 = self.x[idx, :]
                p0 = self.p[idx, :]
                q0 = self.q[idx, :]
                w0 = self.w[idx]
                l0 = self.lam[idx, :]
                b0 = self.beta[idx] * self.beta_decay

                # make a random vector and normalize to get a random direction
                move = torch.empty_like(x0).normal_()
                move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

                # move the cells so that the center of mass of each pair is at the same place as the mother cell was
                x0 = x0 - move/2
                self.x[idx, :] += move/2

                # divide WNT from mother cells evenly to daughter cells
                self.w[idx] /= 2
                w0 /= 2

                # append new cell data to the system state
                self.x = torch.cat((self.x, x0))
                self.p = torch.cat((self.p, p0))
                self.q = torch.cat((self.q, q0))
                self.w = torch.cat((self.w, w0))
                self.lam = torch.cat((self.lam, l0))
                self.beta = torch.cat((self.beta, b0))

        # replenish WNT at source cell(s)
        self.w[self.wnt_cells] = 1

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return division

    def cell_division_single(self):
        """
        Selects exactly one cell to divide and divides it.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False

        idx = torch.multinomial(self.beta, 1)
        with torch.no_grad():
            x0 = self.x[idx, :]
            p0 = self.p[idx, :]
            q0 = self.q[idx, :]
            w0 = self.w[idx]
            l0 = self.lam[idx, :]
            b0 = self.beta[idx] * self.beta_decay

            # make a random vector and normalize to get a random direction
            move = torch.empty_like(x0).normal_()
            move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

            # move the cells so that the center of mass of each pair is at the same place as the mother cell was
            x0 = x0 - move/2
            self.x[idx, :] += move/2

            # divide WNT from mother cells evenly to daughter cells
            self.w[idx] /= 2
            w0 /= 2

            # append new cell data to the system state
            self.x = torch.cat((self.x, x0))
            self.p = torch.cat((self.p, p0))
            self.q = torch.cat((self.q, q0))
            self.w = torch.cat((self.w, w0))
            self.lam = torch.cat((self.lam, l0))
            self.beta = torch.cat((self.beta, b0))

        # replenish WNT at source cell(s)
        self.w[self.wnt_cells] = 1

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return True

def run_and_save(sim, runner, timesteps, yield_every, max_cells, fname_out):
    # Running the simulation
    data = []  # For storing data
    i = 0
    t1 = time.time()
    print('Starting')

    for line in itertools.islice(runner, timesteps):
        i += 1
        print(
            f'Running {i} of {timesteps}   ({yield_every * i} of {yield_every * timesteps})   ({len(line[0])} cells)')
        data.append(line)
        df, lig, kwargs = build_df(data, sim.__dict__)

        if len(line[0]) > max_cells:
            print('Stopping')
            break

        with open(fname_out+'-in-progress.pkl', 'wb') as f:
            pickle.dump({'df':df, 'lig':lig, 'kwargs':kwargs}, f)

    time.sleep(1)

    with open(fname_out+f'{time.strftime("%d%b%Y-%H-%M-%S")}.pkl', 'wb') as f:
        df, lig, kwargs = build_df(data, sim.__dict__)
        pickle.dump({'df':df, 'lig':lig, 'kwargs':kwargs}, f)

    print(f'Simulation done, saved {timesteps} datapoints')
    print('Took', time.time() - t1, 'seconds')


import numpy as np
import pandas as pd
import os
import pickle


def load(fname):
    if fname == "most recent":
        max_mtime = 0
        for dirname, subdirs, files in os.walk("./data/"):
            for file in files:
                full_path = os.path.join(dirname, file)
                mtime = os.stat(full_path).st_mtime
                if mtime > max_mtime:
                    max_mtime = mtime
                    max_file = full_path
        fname = max_file
    print('Loading data from file '+fname)
    with open(fname, 'rb') as f:
        contents = pickle.load(f)
        try:
            data, kwargs = contents
        except ValueError:
            data = contents  # contains x, p, q, lam
            kwargs = None
    # data[t][0] == x, x[i, k] = position of particle i in dimension k
    # data[t][1] == p, p[i, k] = AB polarity of particle i in dimension k
    # data[t][2] == q, q[i, k] = PCP of particle i in dimension k
    return data, kwargs, fname

def build_df(data, kwargs=None, skipframes=1):
    """
    case switcher to inspect data and choose the appropriate dataframe constructor
    it will always return a 3-tuple of values, but the middle one may be None (if the simulation had no ligand information)
    
    INPUT:
        data = a list of tuples containing simulation data, should come from the output of load()
        kwargs = a dictionary of keyword arguments, also from load

    OUTPUT:
        df = pandas DataFrame with columns named appropriately to the data that were passed in
        L or df_lig = 3d numpy array or separate pandas DataFrame defining the ligand distribution, either as a 3d grid of concentration values or a (T*Mx3) dataframe of particle positions
                      if the data did not have ligand information, this return value will be None
        kwargs = pass-through of the kwargs that were inputted
    """
    if len(data)==0:
        raise ValueError('No data to collect')

    if isinstance(data, dict):
        return build_df_from_dict(data, kwargs, skipframes)
    elif len(data[0])==4:
        # data came from a Polar instance
        # variables are x, p, q, lam
        return build_df_plain(data, kwargs, skipframes)
    elif len(data[0])==5:
        # data came from a PolarWNT instance
        # variables are x, p, q, w, lam
        return build_df_wnt(data, kwargs, skipframes)
    elif len(data[0])==6:
        # data came from either a PolarPDE instance or a PolarPattern instance without the 'counts' option
        # to differentiate check the shape of the final entry
        if data[0][-1].ndim==3:
            return build_dfs_ligand_grid(data, kwargs, skipframes)
        else:
            return build_dfs_wnt_ligand(data, kwargs, skipframes)
    elif len(data[0])==7:
        # data came from a PolarPattern instance, including 'counts'
        # variables are x, p, q, w, lam, L, counts
        return build_dfs_wnt_ligand_counts(data, kwargs, skipframes)

def build_df_from_dict(data, kwargs=None, skipframes=1):
    df = data['df']
    lig = data['lig']
    if kwargs in data.keys():
        kwargs = data['kwargs']
    if skipframes > 1:
        df = df.loc[df.t.isin(pd.unique(df.t)[0:-1:skipframes])]
        if isinstance(lig, pd.DataFrame):
            lig = lig.loc[lig.t.isin(pd.unique(lig.t)[0:-1:skipframes])]
        elif isinstance(lig, np.ndarray):
            lig = lig[0:-1:skipframes]
    return df, lig, kwargs

    

def build_df_plain(data, kwargs=None, skipframes=1):
    # create dataframe
    row_chunks = list()
    for t, dat in enumerate(data):
        if t%skipframes == 0:
            if kwargs is not None:
                T = kwargs['dt'] * kwargs['yield_every'] * t
            else:
                T = t
            n = dat[0].shape[0]
            row_chunks.append(np.hstack(
                [np.ones((n, 1)) * T, np.arange(n)[:, np.newaxis], dat[0], dat[1], dat[2]]))

    df = pd.DataFrame(np.vstack(row_chunks), columns=[
                      't', 'i', 'x1', 'x2', 'x3', 'p1', 'p2', 'p3', 'q1', 'q2', 'q3'])
    return df, None, kwargs

def build_df_wnt(data, kwargs=None, skipframes=1):
    row_chunks = list()
    for t, dat in enumerate(data):
        if t % skipframes == 0:
            if kwargs is not None:
                T = kwargs['dt'] * kwargs['yield_every'] * t
            else:
                T = t
            n = dat[0].shape[0]
            row_chunks.append(np.hstack([np.ones(
                (n, 1)) * T, np.arange(n)[:, np.newaxis], dat[0], dat[1], dat[2], dat[3][:, None]]))

    df = pd.DataFrame(np.vstack(row_chunks), columns=[
                      't', 'i', 'x1', 'x2', 'x3', 'p1', 'p2', 'p3', 'q1', 'q2', 'q3', 'w'])
    return df, None, kwargs

def build_dfs_wnt_ligand(data, kwargs=None, skipframes=1):
    row_chunks = list()
    ligand_chunks = list()
    for t, dat in enumerate(data):
        if t % skipframes == 0:
            if kwargs is not None:
                T = kwargs['dt'] * kwargs['yield_every'] * t
            else:
                T = t
            n = dat[0].shape[0]
            m = dat[5].shape[0]
            row_chunks.append(np.hstack([np.ones(
                (n, 1)) * T, np.arange(n)[:, np.newaxis], dat[0], dat[1], dat[2], dat[3][:, None]]))
            ligand_chunks.append(
                np.hstack([np.ones((m, 1)) * T, np.arange(m)[:, np.newaxis], dat[5]]))

    df = pd.DataFrame(np.vstack(row_chunks), columns=[
                      't', 'i', 'x1', 'x2', 'x3', 'p1', 'p2', 'p3', 'q1', 'q2', 'q3', 'w'])
    df_lig = pd.DataFrame(np.vstack(ligand_chunks), columns=[
                          't', 'j', 'x1', 'x2', 'x3'])
    return df, df_lig, kwargs

def build_dfs_ligand_grid(data, kwargs=None, skipframes=1):
    row_chunks = list()
    ligand_chunks = list()
    for t, dat in enumerate(data):
        if t % skipframes == 0:
            if kwargs is not None:
                T = kwargs['dt'] * kwargs['yield_every'] * t
            else:
                T = t
            n = dat[0].shape[0]
            m = dat[5].shape[0]
            row_chunks.append(np.hstack([np.ones(
                (n, 1)) * T, np.arange(n)[:, np.newaxis], dat[0], dat[1], dat[2], dat[3][:, None]]))
            ligand_chunks.append(dat[5])

    df = pd.DataFrame(np.vstack(row_chunks), columns=[
                      't', 'i', 'x1', 'x2', 'x3', 'p1', 'p2', 'p3', 'q1', 'q2', 'q3', 'w'])
    L = np.stack(ligand_chunks)
    return df, L, kwargs

def build_dfs_wnt_ligand_counts(data, kwargs=None, skipframes=1):
    row_chunks = list()
    ligand_chunks = list()
    for t, dat in enumerate(data):
        if t % skipframes == 0:
            if kwargs is not None:
                T = kwargs['dt'] * kwargs['yield_every'] * t
            else:
                T = t
            n = dat[0].shape[0]
            m = dat[5].shape[0]
            counts_padded = np.pad(dat[6], (0, n-len(dat[6])), mode='constant', constant_values = (0,0))
            row_chunks.append(np.hstack([np.ones(
                (n, 1)) * T, np.arange(n)[:, np.newaxis], dat[0], dat[1], dat[2], dat[3][:, None], counts_padded[:, None]]))
            ligand_chunks.append(
                np.hstack([np.ones((m, 1)) * T, np.arange(m)[:, np.newaxis], dat[5]]))

    df = pd.DataFrame(np.vstack(row_chunks), columns=[
                      't', 'i', 'x1', 'x2', 'x3', 'p1', 'p2', 'p3', 'q1', 'q2', 'q3', 'w', 'count'])
    df_lig = pd.DataFrame(np.vstack(ligand_chunks), columns=[
                          't', 'j', 'x1', 'x2', 'x3'])
    return df, df_lig, kwargs


def select(df, T_plot, kwargs=None):
    if T_plot == -1:
        tt = df['t'].max()
    else:
        tt = df.loc[np.argmin((df['t']-T_plot)**2), 't']
    mask = df['t'] == tt
    return df[mask].copy()



import torch


def potential_nematic(x, d, dx, lam_i, lam_j, pi, pj, qi, qj, Gi):
    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.abs(torch.sum(torch.cross(pi, qi, dim=2) *
                   torch.cross(pj, qj, dim=2), dim=2))
    S3 = torch.abs(torch.sum(torch.cross(qi, dx, dim=2) *
                   torch.cross(qj, dx, dim=2), dim=2))
    S4 = torch.sum(torch.cross(qi, Gi, dim=2) *
                   torch.cross(qi, Gi, dim=2), dim=2)

    # Combination must be symmetric to abide by Newton's 3rd Law
    lam = (lam_i + lam_j) / 2
    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    # Calculate semi-nematic potential
    S = lam[:, :, 0] + lam[:, :, 1] * S1 + \
        lam[:, :, 2] * S2 + lam[:, :, 3] * S3 + lam[:, :, 4] * S4
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij


def potential_nematic_reweight(x, d, dx, lam_i, lam_j, pi, pj, qi, qj, Gi):
    """
    Implements the nematic potential and re-weights lambda at cells whose WNT is below threshold

    to detect which cells have WNT level below threshold, check the norm of Gi
    """
    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.abs(torch.sum(torch.cross(pi, qi, dim=2) *
                   torch.cross(pj, qj, dim=2), dim=2))
    S3 = torch.abs(torch.sum(torch.cross(qi, dx, dim=2) *
                   torch.cross(qj, dx, dim=2), dim=2))
    S4 = torch.sum(torch.cross(qi, Gi, dim=2) *
                   torch.cross(qi, Gi, dim=2), dim=2)

    # Combination must be symmetric to abide by Newton's 3rd Law
    lam = (lam_i + lam_j) / 2
    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]
    reweight_mask = ((Gi**2).sum(axis=2).mean(axis=1) < 0.999)
    lam[reweight_mask, :, :] /= lam[reweight_mask, :, :-1].sum(axis=-1)[:, :, None].expand(reweight_mask.sum(), lam.shape[1], lam.shape[2])

    # Calculate semi-nematic potential
    S = lam[:, :, 0] + lam[:, :, 1] * S1 + \
        lam[:, :, 2] * S2 + lam[:, :, 3] * S3 + lam[:, :, 4] * S4
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij


def potential_vectorial(x, d, dx, lam_i, lam_j, pi, pj, qi, qj, Gi):
    S1 = torch.sum(torch.cross(pj, dx, dim=2) *
                   torch.cross(pi, dx, dim=2), dim=2)
    S2 = torch.sum(torch.cross(pi, qi, dim=2) *
                   torch.cross(pj, qj, dim=2), dim=2)
    S3 = torch.sum(torch.cross(qi, dx, dim=2) *
                   torch.cross(qj, dx, dim=2), dim=2)
    S4 = torch.sum(torch.cross(qi, Gi, dim=2) *
                   torch.cross(qi, Gi, dim=2), dim=2)

    # Combination must be symmetric to abide by Newton's 3rd Law
    lam = (lam_i + lam_j) / 2
    lam1 = 0.5 * (lam_i + lam_j)
    lam2 = lam1.clone()
    lam2[:, : 0] = 1
    lam2[:, :, 1:] = 0
    mask1 = 1 * (lam1[:, :, 0] > 0.5)

    lam = lam1 * (1 - mask1[:, :, None]) + lam2 * mask1[:, :, None]

    # Calculate semi-nematic potential
    S = lam[:, :, 0] + lam[:, :, 1] * S1 + \
        lam[:, :, 2] * S2 + lam[:, :, 3] * S3 + lam[:, :, 4] * S4
    Vij = torch.exp(-d) - S * torch.exp(-d / 5)
    return Vij




import os
from .plotcore import load, build_df, select

import pandas as pd
import numpy as np
import plotly.express as px

def good_basis(normal_vector):
    nv = normal_vector/np.linalg.norm(normal_vector)
    v2 = np.cross(nv, np.random.rand(3))
    v2 /= np.linalg.norm(v2)
    v3 = np.cross(normal_vector, v2)
    v3 /= np.linalg.norm(v3)
    V = np.array([v2, v3]).T
    return nv, V

def plot(df, normal_vector, offset, width):
    nv, V = good_basis(normal_vector)
    X_all = df[['x1','x2','x3']].values
    normal_values = X_all @ nv
    mask_left = (normal_values > offset - width)
    mask_right = (normal_values < offset + width)
    X_slice = X_all[mask_left * mask_right]
    data2d = X_slice @ V
    fig = px.density_heatmap(pd.DataFrame(data2d), x=0, y=1, nbinsx=100, nbinsy=100)
    return fig
    

def save(fig, fname, T_plot):
    fig.write_html(os.path.join('animations', os.path.basename(fname)).replace('.pkl',f'_ligand_heatmap_t-{T_plot}.html'), include_plotlyjs='directory',
                   full_html=False)
    return

if __name__ == '__main__':
    fname = input('Enter data filename (default: most recent): ') or 'most recent' # 'data/test1.pkl'
    T_plot = int(input('timestep to plot (default: final time): ') or -1)
    vector_input = input('normal vector? (default = [1,0,0]) ') or '1,0,0'
    offset = float(input('offset? (default: 0) ') or 0)
    width = float(input('width? (default: 3)') or 3)
    normal_vector = np.array(eval(vector_input), dtype = 'float')
    data, kwargs, fname = load(fname)
    _, df_lig, kwargs = build_df(data, kwargs)
    df = select(df_lig, T_plot)
    fig = plot(df, normal_vector, offset, width)
    save(fig, fname, T_plot)



import os
import plotly.express as px
from .plotcore import load, build_df



def plot(df, color = 'x1'):
    range_x = (df['x1'].min(), df['x1'].max())
    range_y = (df['x2'].min(), df['x2'].max())
    range_z = (df['x3'].min(), df['x3'].max())
    fig = px.scatter_3d(df, x='x1', y='x2', z='x3', animation_frame='t',
                        color = color, range_x=range_x, range_y=range_y, range_z=range_z)

    def fun(scene):
        scene.aspectmode = 'data'
        return
    fig.for_each_scene(fun)
    return fig


def save(fig, fname):
    fig.write_html(os.path.join('animations', os.path.basename(fname)).replace('.pkl','_ligand.html'), include_plotlyjs='directory',
                   full_html=False, animation_opts={'frame': {'duration': 50}})


if __name__ == '__main__':
    fname = input('Enter data filename (default: most recent): ') or 'most recent' # 'data/test1.pkl'
    skipframes = input('plot ever n frames? (default: 1) ') or 1
    color = input('color by? (default: x1) ') or 'x1'
    data, kwargs, fname = load(fname)
    _, df_lig, kwargs = build_df(data, kwargs, skipframes = int(skipframes))
    fig = plot(df_lig, color=color)
    save(fig, fname)



"""
This file defines the class PolarGrid, which implements the polarity model coupled to a PDE describing the evolution of ligand density around the growing organoid

Be very careful with grid spacing!!!! Number of variables grows with the cube of the inverse grid spacing.
"""
from .polarcore import PolarWNT
import numpy as np
import torch


class PolarPDE(PolarWNT):
    def __init__(self, *args,
                 bounding_radius_factor=2,
                 contact_radius=1,
                 R_init=1.0,
                 a=0.8,
                 gamma=0.1,
                 b=1.0,
                 D=100,
                 hill_coefficient=8,
                 hill_k=0.55,
                 grid_N=100,
                 grid_dx=0.5,
                 beta_func_of_w_exponent=1,
                 **kwargs):
        self.bounding_radius_factor = bounding_radius_factor
        self.contact_radius = contact_radius
        super().__init__(*args, **kwargs)
        self.R = (2*torch.rand(*self.w.shape, requires_grad=False,
                  device=self.device)) * R_init
        self.R.requires_grad = False
        self.R_init = R_init
        # intrinsic receptor production rate (as a multiple of gamma)
        self.a = a
        # intrinsic ligand production rate (as a multiple of gamma)
        self.b = b
        self.gamma = gamma                  # intrinsic decay rate of receptor
        self.D = D                          # diffusion ratio of ligand vs. receptor
        self.grid_N = grid_N                # number of gridpoints in each dimension
        self.grid_dx = grid_dx              # spatial distance between adjacent gridpoints
        self.beta_func_of_w_exponent = beta_func_of_w_exponent
        # initialize a 3d tensor to hold the ligand level in space
        self.L = torch.zeros(
            (self.grid_N, self.grid_N, self.grid_N), device=self.device)
        self.stencil_3d_conv = torch.tensor([[[2, 3, 2], [3, 6, 3], [2, 3, 2]],
                                             [[3, 6, 3], [6, -88, 6], [3, 6, 3]],
                                             [[2, 3, 2], [3, 6, 3], [2, 3, 2]]], device=self.device) / (26*self.grid_dx**2)
        # create a tensor giving spatial positions corresponding to indices of self.L
        Lcoords = self.grid_dx * \
            (torch.arange(self.grid_N, device=self.device) - (self.grid_N - 1)/2)
        self.L_grid = torch.stack(torch.meshgrid(
            (Lcoords, Lcoords, Lcoords), indexing='ij'), dim=-1)
        self.hill_coefficient = hill_coefficient
        self.hill_k = hill_k
        self.get_bounding_sphere()

    def get_bounding_sphere(self):
        self.bounding_sphere_center = self.x.mean(dim=0).detach()
        self.bounding_sphere_radius = (self.bounding_radius_factor *
                                       torch.max(torch.norm(self.x - self.bounding_sphere_center, dim=1))).detach()
        return self.bounding_sphere_center, self.bounding_sphere_radius

    def hill_function(self, x):
        return x**self.hill_coefficient / (self.hill_k**self.hill_coefficient + x**self.hill_coefficient)

    def laplacian_R(self):
        """
        Computes the laplacian of the R concentration field
        computes the graph laplacian according to "true neighbors" including nearest-neighbor distances 
        normalization is tricky since neighbor-pairs are not in orthogonal directions...
            i think this means we should normalize by 4/(# neighbors)
        """
        self.find_true_neighbours()
        # return ((self.R[self.idx]- self.idx.shape[1]*self.R[:, None]) / self.d**2).sum(dim=1) * (4/self.idx.shape[1])
        return ((self.R[self.idx] - self.idx.shape[1]*self.R[:,None])/self.d.detach()**2).sum(dim=1) * (4/self.idx.shape[1])

    def laplacian_L(self):
        """
        computes the laplacian of the continuous field L, which is defined on a 3D rectangular grid with uniform spacing dx

        math formula: laplacian_x[i,j,k] = ((x[i+1,j,k] + x[i-1,j,k] + x[i,j+1,k] + x[i,j-1,k] + x[i,j,k+1] + x[i,j,k-1]) - 6 * x[i,j,k])/x**2
        This holds for i,j,k in the /interior/ of the cube; if any of i,j,k +/- 1 is not in the domain, replace it with i,j,k

        I want to do this with clever indexing and padding
        """
        A = torch.cat((self.L[1:, :, :], self.L[-1:, :, :]),
                      dim=0)  # A[i,j,k] = self.L[i+1,j,k]
        # B[i,j,k] = self.L[i-1,j,k]
        B = torch.cat((self.L[:1, :, :], self.L[:-1, :, :]), dim=0)
        # C[i,j,k] = self.L[i,j+1,k]
        C = torch.cat((self.L[:, 1:, :], self.L[:, -1:, :]), dim=1)
        # D[i,j,k] = self.L[i,j-1,k]
        D = torch.cat((self.L[:, :1, :], self.L[:, :-1, :]), dim=1)
        # E[i,j,k] = self.L[i,j,k+1]
        E = torch.cat((self.L[:, :, 1:], self.L[:, :, -1:]), dim=2)
        # F[i,j,k] = self.L[i,j,k-1]
        F = torch.cat((self.L[:, :, :1], self.L[:, :, :-1]), dim=2)
        return (A+B+C+D+E+F - 6*self.L)/(self.grid_dx**2)

    def laplacian_L_27_point(self):
        """
        Implements the 27-point laplacian stencil, which should be more numerically stable than the 7-point one above
        """
        # first pad L with identical data in each dimension
        Lpad = torch.nn.functional.pad(self.L[None,None,:,:,:], (1,1,1,1,1,1), 'replicate')
        return torch.nn.functional.conv3d(Lpad, self.stencil_3d_conv[None, None, :, :])[0, 0]

    def closest_gridpoints(self):
        """
        Finds closest gridpoint to cell locations

        NOTE: Relies on the spatial locations are defined in the same way in each dimension, i.e. the spatial grid is cubic and does not change
        In particular index i corresponds to spatial location (i-(N-1)/2)*dx

        INPUT:
            self.x = tensor of shape (N, 3) containing (x,y,z)-coordinates of N points, to which we want the closest gridpoint
            self.grid_dx = spatial distance between grid points along each dimension
            self.grid_N = number of spatial grid points in each dimension

        OUTPUT:
            idx = tuple of tensors, each of length N
                  index into a 3D tensor by doing L[idx]
        """
        # check if any coordinates are outside
        if self.x.max() > self.grid_dx*(self.grid_N - 1)/2 or self.x.min() < -self.grid_dx*(self.grid_N - 1)/2:
            raise RuntimeError('cell positions out of bounds')
        idxs_2d = torch.round(self.x/self.grid_dx +
                              (self.grid_N - 1)/2).long().chunk(chunks=3, dim=1)
        return tuple(ind.squeeze() for ind in idxs_2d)

    def L_source_indices(self):
        """
        This routine finds the 3-dimensional indices of gridpoints where L should be produced.
        At the moment this will be  all gridpoints OUTSIDE of the bouding sphere

        This may end up capturing a lot of compute time, could be a place to look in the future to speed up by figuring out some clever direct way to get these indices
        To do that would probably require parametrizing the bounding sphere somehow
        """
        self.get_bounding_sphere()
        return torch.where((self.L_grid - self.bounding_sphere_center).norm(dim=-1) > self.bounding_sphere_radius)

    def reaction_diffusion_dynamics(self):
        """
        This is the routine that computes the update to L and to R
        The dynamics are:
            Epithelium:
                dR/dt = Laplacian(R) + gamma*(a - R + R^2 * L)
                dL/dt = D*Laplacian(L) + gamma*(-R^2 * L)
            Mesenchyme:
                dL/dt = D*Laplacian(L) + gamma*(b - L)

        For me, the term -gamma*L in dL/dt is active everywhere EXCEPT at those gridpoints in contact with cells.
        However, the term +gamma*b is only active at those gridpoints specified by self.L_source_indices
        """
        idx = self.closest_gridpoints() # these indices correspond to the epithelium
        dL = self.dt * (self.D * self.laplacian_L_27_point() - self.gamma * self.L)
        dL[idx] += self.dt * self.gamma * ((-self.R**2) * self.L[idx])
        dL[self.L_source_indices()] += self.dt * self.gamma * self.b
        # dL += self.dt * self.gamma * self.b

        dR = self.dt * (self.laplacian_R()*0 + self.gamma *
                        (self.a - self.R + self.hill_function(self.R**2 * self.L[idx])))

        self.L += dL
        self.R += dR

    def simulation(self, potential,
                   division_decider=lambda *args: False,
                   better_WNT_gradient=False,
                   wnt_func_of_R=False,
                   beta_func_of_w=False):
        """
        Generator to implement the simulation

        Note: you can interact with this thing. Example:
        ```python
        polarguy = PolarWNT(**)
        sim = polarguy.simulation(***)
        for out in sim:
            data.append(out)
            polarguy.get_gradient_averaging()
        ```

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            eta : float
                Strength of the added noise
            potential : callable
                function that computes the value of the potential between two cells, i and j
                call signature (x, d, dx, lam_i, lam_j, pi, pj, qi, qj)
            yield_every : int, optional
                How many simulation time steps to take between yielding the system state. Default: 1
            dt : float, optional
                Size of the time step. Default: 0.1
            kwargs : dict
                Keyword args passed to self.init_simulation and self.time_step.
                Values passed here override default values
                dt : float
                    time step
                yield_every : int
                    how many time steps to take in between yielding system state

        Yields
        ----------
            x : numpy.ndarray
                Position of each cell in 3D space
            p : numpy.ndarray
                AB polarity vector of each cell
            q : numpy.ndarray
                PCP vector of each cell
            w : numpy.ndarray
                WNT concentration at each cell
            lam : numpy.ndarray
                weights for the terms of the potential function for each cell.
        """

        tstep = 0
        while True:
            tstep += 1

            # perform cell division, depending on the output of the function division_decider
            # by default, always do cell division (this results in exponential growth of number of cells)
            # decide if cells divide this time
            division = False
            if division_decider(self, tstep):
                if self.divide_single:
                    division = self.cell_division_single()
                else:
                    division = self.cell_division()

            n_update = self.update_k(self.true_neighbour_max, tstep)
            self.k = min(self.k, len(self.x) - 1)

            # recompute potential neighbors if necessary
            if division or tstep % n_update == 0 or self.idx is None:
                self.find_potential_neighbours()

            # do the random walking of ligand particles
            # this is done _after_ cell division so the contact_counts array always has the same size as the cell position array
            self.reaction_diffusion_dynamics()

            self.get_gradient_vectors(better=better_WNT_gradient)
            self.gradient_step(tstep, potential=potential)

            if wnt_func_of_R:
                self.w = self.hill_function(self.R)
            else:
                self.w = self.w * np.exp(self.dt * self.wnt_decay)
            if beta_func_of_w:
                www = self.w.detach().clone()
                self.beta = ((www - www.min())/(www.max() - www.min())) ** self.beta_func_of_w_exponent

            if tstep % self.yield_every == 0:
                print(self.R.min(), self.R.max(),
                      (self.R.max() - self.R.min()))
                print(self.L.max(), self.L.min(), (self.L.max()-self.L.min()))
                xx = self.x.detach().to("cpu").numpy().copy()
                pp = self.p.detach().to("cpu").numpy().copy()
                qq = self.q.detach().to("cpu").numpy().copy()
                ww = self.w.detach().to("cpu").numpy().copy()
                ll = self.lam.detach().to("cpu").numpy().copy()
                lig = self.L.detach().to('cpu').numpy().copy()
                yield xx, pp, qq, ww, ll, lig

    def cell_division(self):
        """
        Decides which cells divide, and if they do, places daughter cells.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction

        Parameters
        ----------
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
            dt : float
                Size of the time step.
            kwargs : dict
                Valid keyword arguments:
                beta_decay : float
                    the factor by which beta (probability of cell division per unit time) decays upon cell division.
                    after cell division, one daughter cell has the same beta as the mother (b0), and the other has beta = b0 * beta_decay

        Returns
        ---------
            division : bool
                True if cell division has taken place, otherwise False
            x : torch.Tensor
                Position of each cell in 3D space
            p : torch.Tensor
                AB polarity vector of each cell
            q : torch.Tensor
                PCP vector of each cell
            lam : torch.Tensor
                weights for the terms of the potential function for each cell.
            beta : torch.Tensor
                for each cell, probability of division per unit time
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False

        # set probability according to beta and dt
        d_prob = self.beta * self.dt
        # flip coins
        draw = torch.empty_like(self.beta).uniform_()
        # find successes
        events = draw < d_prob
        division = False

        if torch.sum(events) > 0:
            with torch.no_grad():
                division = True
                # find cells that will divide
                idx = torch.nonzero(events)[:, 0]

                x0 = self.x[idx, :]
                p0 = self.p[idx, :]
                q0 = self.q[idx, :]
                w0 = self.w[idx]
                R0 = self.R[idx]
                l0 = self.lam[idx, :]
                b0 = self.beta[idx] * self.beta_decay

                # make a random vector and normalize to get a random direction
                move = torch.empty_like(x0).normal_()
                move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

                # move the cells so that the center of mass of each pair is at the same place as the mother cell was
                x0 = x0 - move/2
                self.x[idx, :] += move/2

                # divide WNT from mother cells evenly to daughter cells
                self.w[idx] /= 2
                w0 /= 2
                self.R[idx] /= 2
                R0 /= 2

                # append new cell data to the system state
                self.x = torch.cat((self.x, x0))
                self.p = torch.cat((self.p, p0))
                self.q = torch.cat((self.q, q0))
                self.w = torch.cat((self.w, w0))
                self.R = torch.cat((self.R, R0))
                self.lam = torch.cat((self.lam, l0))
                self.beta = torch.cat((self.beta, b0))

        # replenish WNT at source cell(s)
        self.w[self.wnt_cells] = 1

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return division

    def cell_division_single(self):
        """
        Selects exactly one cell to divide and divides it.
        If a cell divides, one daughter cell is placed at the same position as the parent cell, and the other is placed one cell diameter away in a uniformly random direction
        """
        if torch.sum(self.beta) < self.do_nothing_threshold:
            return False

        idx = torch.multinomial(self.beta, 1)
        with torch.no_grad():
            x0 = self.x[idx, :]
            p0 = self.p[idx, :]
            q0 = self.q[idx, :]
            w0 = self.w[idx]
            R0 = self.R[idx]
            l0 = self.lam[idx, :]
            b0 = self.beta[idx] * self.beta_decay

            # make a random vector and normalize to get a random direction
            move = torch.empty_like(x0).normal_()
            move /= torch.sqrt(torch.sum(move**2, dim=1))[:, None]

            # move the cells so that the center of mass of each pair is at the same place as the mother cell was
            x0 = x0 - move/2
            self.x[idx, :] += move/2

            # divide WNT from mother cells evenly to daughter cells
            self.w[idx] /= 2
            w0 /= 2
            self.R[idx] /= 2
            R0 /= 2

            # append new cell data to the system state
            self.x = torch.cat((self.x, x0))
            self.p = torch.cat((self.p, p0))
            self.q = torch.cat((self.q, q0))
            self.w = torch.cat((self.w, w0))
            self.R = torch.cat((self.R, R0))
            self.lam = torch.cat((self.lam, l0))
            self.beta = torch.cat((self.beta, b0))

        # replenish WNT at source cell(s)
        self.w[self.wnt_cells] = 1

        self.x.requires_grad = True
        self.p.requires_grad = True
        self.q.requires_grad = True

        return True


"""
This file is to plot the density of ligand in the case that the ligand is modeled by a PDE instead of particles
"""
import os
from .plotcore import load, build_df
import plotly.express as px


def plot(L, axis, index):
    fig = px.imshow(L.take(index, axis), animation_frame=0)
    return fig


def save(fig, fname):
    fig.write_html(os.path.join('animations',os.path.basename(fname)).replace('.pkl', f'_ligand_density_pde.html'),
                   include_plotlyjs='directory',
                   full_html=False)
    return

if __name__ == '__main__':
    fname = input('Enter data filename (default: most recent): ') or 'most recent' # 'data/test1.pkl'
    data, kwargs, fname = load(fname)
    df, L, kwargs = build_df(data, kwargs)
    axis = int(input('axis along which to slice? default 1') or 1)
    index = int(input('which slice to take? (default: halfway) ') or L.shape[-1]/2)
    fig = plot(L, axis, index)
    save(fig, fname)



import setuptools
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="polaritymodel",
    version=get_version("src/polaritymodel/__init__.py"),
    author="Jordan Snyder",
    author_email="snydrew@gmail.com",
    description="A package for running the cell polarity model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jasnyder/polarpkg",
    project_urls={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    install_requires=[
        'torch',
        'torchaudio',
        'torchvision',
        'numpy',
        'scipy',
        'plotly',
        'pandas'
    ],
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)


import os
from .plotcore import load, select, build_df
import plotly.graph_objects as go

def plot(df):
    fig = go.Figure(data=[go.Cone(
        x=df['x1'],
        y=df['x2'],
        z=df['x3'],
        u=df['q1'],
        v=df['q2'],
        w=df['q3'],
        sizemode='absolute',
        sizeref=2
    )])

    def fun(scene):
        scene.aspectmode = 'data'
        return
    fig.for_each_scene(fun)
    return fig

def save(fig, fname):
    fig.write_html(os.path.join('animations', os.path.basename(fname)).replace('.pkl','_vectorfield.html'),
                   include_plotlyjs='directory', full_html=False, animation_opts={'frame': {'duration': 100}})

if __name__ == "__main__":
    fname = input('Enter data filename (default: most recent): ') or 'most recent'
    T_plot = int(input('timestep to plot: ') or -1)
    data, kwargs, fname = load(fname)
    df, _, kwargs = build_df(data, kwargs)
    df_t = select(df, T_plot, kwargs)
    fig = plot(df_t)
    save(fig, fname)